#include "map/map.h"

#include <iostream>

#include "tiles/grass.h"

#define MOD(a, b)    (((b) + ((a) % (b))) % (b))

Node &Map::at(Vector2<int> pos) {
    Vector2<int> region_offset = pos / (REGION_SIZE + 0);
    // Floor division
    if(pos.x() < 0 && pos.x() % REGION_SIZE) region_offset -= {1, 0};
    if(pos.y() < 0 && pos.y() % REGION_SIZE) region_offset -= {0, 1};

    auto it = m_reigons.find(region_offset);

    if(it == m_reigons.end()) {
        m_reigons.insert(
            std::pair<Vector2<int>, Region>(
                region_offset,
                load_or_create_region(region_offset)
            )
        );
    }

    Region &region = m_reigons.at(region_offset);

    Vector2<int> tile_offset = Vector2<int>(MOD(pos.x(), REGION_SIZE), MOD(pos.y(), REGION_SIZE));
    return region.at(tile_offset);
}

void Map::set_tile(Vector2<int> pos, Tile *tile) {
    at(pos).set_tile(tile);
}

void Map::save() {
    for(auto &elm : m_reigons) {
        elm.second.save();
    }
}

void Map::load() {
    for(auto &elm : m_reigons) {
        elm.second.load();
    }
}

Region Map::load_or_create_region(Vector2<int> pos) {
    Region region(pos);
    if(!region.load()) {
        std::cout << "Generating region @" << pos << std::endl;
        // Call generator
        m_grass.generate(pos, region);

        // Save the newly generated region
        region.save();
    } else {
        std::cout << "Loaded region @" << pos << std::endl;
    }
    return region;
}

