#pragma once

#include <iostream>

#include "tiles/tile.h"
#include "entities/entity.h"
#include "util/render.h"
#include "util/vector2.h"


class Node {
    public:
        Node() : Node({0, 0}, {1, 1}) {}
        Node(Vector2<int> pos, Vector2<int> size) :
            m_pos(pos),
            m_size(size) {}

        Node(const Node &other);
        Node &operator=(const Node &other) = default;

        ~Node();

        // Get a reference to the tile associated with this node
        Tile &tile() { return *m_tile; }
        void set_tile(Tile *tile);

        // Get a reference to the tile entity associated with this node
        Entity &entity() { return *m_entity; }
        void set_entity(Entity *tile);

        // Get/Set the position in world coordinates
        Vector2<int> pos() { return m_pos; }
        void set_pos(Vector2<int> pos) { m_pos = pos; }

        // Get/Set the size of this block in tiles
        Vector2<int> size() { return m_size; }
        void set_size(Vector2<int> size) { m_size = size; }

        void update(double dt);
        void render(Render r);

        friend std::ostream &operator<<(std::ostream &os, const Node &node);
        friend std::istream &operator>>(std::istream &is, Node &node);


    private:
        Tile *m_tile = 0;
        Entity *m_entity = 0;

        Vector2<int> m_pos;
        Vector2<int> m_size;
};
