#pragma once

#include <ostream>

template <typename T>
class Vector2 {
    public:
        Vector2() = default;

        Vector2(T x, T y) : m_x(x), m_y(y) {}
        template <typename X, typename Y>
        Vector2(X x, Y y) :
            m_x(static_cast<T>(x)),
            m_y(static_cast<T>(y)) {}

        Vector2(const Vector2<T> &other) = default;
        template <typename A>
        Vector2(const Vector2<A> &other) :
            m_x(static_cast<T>(other[0])),
            m_y(static_cast<T>(other[1])) {};

        T x() { return m_x; }
        T y() { return m_y; }

        void set_x(T x) { m_x = x; }
        void set_y(T y) { m_y = y; }

        void set(T x, T y) { m_x = x; m_y = y; }

        Vector2<T> operator-() const { return Vector2<T>(-m_x, -m_y); }

        Vector2<T> &operator=(const Vector2<T> &other) { m_x = other.m_x; m_y = other.m_y; return *this; }

        Vector2<T> operator+(const Vector2<T> &other) const { return Vector2(m_x + other.m_x, m_y + other.m_y); }
        Vector2<T> &operator+=(const Vector2<T> &other) { m_x += other.m_x; m_y += other.m_y; return *this; }

        Vector2<T> operator-(const Vector2<T> &other) const { return Vector2(m_x - other.m_x, m_y - other.m_y); }
        Vector2<T> &operator-=(const Vector2<T> &other) { m_x -= other.m_x; m_y -= other.m_y; return *this; }

        Vector2<T> operator*(const Vector2<T> &other) const { return Vector2(m_x * other.m_x, m_y * other.m_y); }
        Vector2<T> &operator*=(const Vector2<T> &other) { m_x *= other.m_x; m_y *= other.m_y; return *this; }
        Vector2<T> operator*(const T &s) const { return Vector2(m_x * s, m_y * s); }
        Vector2<T> &operator*=(const T &s) { m_x *= s; m_y *= s; return *this; }

        Vector2<T> operator/(const Vector2<T> &other) const { return Vector2(m_x / other.m_x, m_y / other.m_y); }
        Vector2<T> &operator/=(const Vector2<T> &other) { m_x /= other.m_x; m_y /= other.m_y; return *this; }
        Vector2<T> operator/(const T &s) const { return Vector2(m_x / s, m_y / s); }
        Vector2<T> &operator/=(const T &s) { m_x /= s; m_y /= s; return *this; }

        bool operator==(const Vector2<T> &other) const { return m_x == other.m_x && m_y == other.m_y; }
        bool operator!=(const Vector2<T> &other) const { return m_x != other.m_x || m_y != other.m_y; }

        bool operator<(const Vector2<T> &other) const { return m_x < other.m_x || (m_x == other.m_x && m_y < other.m_y); }
        bool operator>(const Vector2<T> &other) const { return m_x > other.m_x || (m_x == other.m_x && m_y > other.m_y); }

        bool operator<=(const Vector2<T> &other) const { return m_x <= other.m_x || (m_x == other.m_x && m_y <= other.m_y); }
        bool operator>=(const Vector2<T> &other) const { return m_x >= other.m_x || (m_x == other.m_x && m_y >= other.m_y); }

        T &operator[](const int idx) {
            switch(idx) {
                default:
                case 0:
                    return m_x;
                case 1:
                    return m_y;
            }
        }
        const T &operator[](const int idx) const {
            switch(idx) {
                default:
                case 0:
                    return m_x;
                case 1:
                    return m_y;
            }
        }

        friend std::ostream &operator<<(std::ostream &os, const Vector2<T> &vec) {
            os << "Vector2(" << vec.m_x << ", " << vec.m_y << ")";
            return os;
        }

    private:
        T m_x = 0;
        T m_y = 0;
};

